/*
 * Windows Shell Item extension block functions
 *
 * Copyright (c) 2010-2013, Joachim Metz <joachim.metz@gmail.com>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include "libfwsi_definitions.h"
#include "libfwsi_extension_block.h"
#include "libfwsi_file_attributes.h"
#include "libfwsi_folder_identifier.h"
#include "libfwsi_libcerror.h"
#include "libfwsi_libcnotify.h"
#include "libfwsi_libcstring.h"
#include "libfwsi_libfdatetime.h"
#include "libfwsi_libfguid.h"
#include "libfwsi_libuna.h"
#include "libfwsi_types.h"

/* Creates an extension block
 * Make sure the value extension_block is referencing, is set to NULL
 * Returns 1 if successful or -1 on error
 */
int libfwsi_extension_block_initialize(
     libfwsi_extension_block_t **extension_block,
     libcerror_error_t **error )
{
	static char *function = "libfwsi_extension_block_initialize";

	if( extension_block == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid extension block.",
		 function );

		return( -1 );
	}
	if( *extension_block != NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid extension block value already set.",
		 function );

		return( -1 );
	}
	*extension_block = memory_allocate_structure(
	                    libfwsi_extension_block_t );

	if( *extension_block == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create extension block.",
		 function );

		goto on_error;
	}
	if( memory_set(
	     *extension_block,
	     0,
	     sizeof( libfwsi_extension_block_t ) ) == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear extension block.",
		 function );

		goto on_error;
	}
	return( 1 );

on_error:
	if( extension_block != NULL )
	{
		memory_free(
		 *extension_block );

		*extension_block = NULL;
	}
	return( -1 );
}

/* Frees an extension block
 * Returns 1 if successful or -1 on error
 */
int libfwsi_extension_block_free(
     libfwsi_extension_block_t **extension_block,
     libcerror_error_t **error )
{
	static char *function = "libfwsi_extension_block_free";

	if( extension_block == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid extension block.",
		 function );

		return( -1 );
	}
	if( *extension_block != NULL )
	{
		memory_free(
		 *extension_block );

		*extension_block = NULL;
	}
	return( 1 );
}

/* Copies a shell extension block from a byte stream
 * Returns 1 if successful or -1 on error
 */
int libfwsi_extension_block_copy_from_byte_stream(
     libfwsi_extension_block_t *extension_block,
     const uint8_t *byte_stream,
     size_t byte_stream_size,
     int ascii_codepage,
     libcerror_error_t **error )
{
	uint8_t curi_class_identifier[ 16 ]         = { 0x13, 0xce, 0x2f, 0xdf, 0xec, 0x25, 0xbb, 0x45, 0x9d, 0x4c, 0xce, 0xcd, 0x47, 0xc2, 0x43, 0x0c };
	static char *function                       = "libfwsi_extension_block_copy_from_byte_stream";
	size_t extension_block_data_size            = 0;
	size_t string_size                          = 0;
	uint32_t number_of_properties               = 0;
	uint32_t property_index                     = 0;
	uint32_t property_size                      = 0;
	uint32_t property_type                      = 0;
	uint16_t long_string_size                   = 0;
	int result                                  = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	libcstring_system_character_t date_time_string[ 32 ];
        libcstring_system_character_t guid_string[ 48 ];

	libcstring_system_character_t *value_string = NULL;
	libfdatetime_fat_date_time_t *fat_date_time = NULL;
        libfguid_identifier_t *guid                 = NULL;
	size_t value_string_size                    = 0;
	uint32_t value_32bit                        = 0;
	uint16_t value_16bit                        = 0;
#endif

	if( extension_block == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid extension block.",
		 function );

		return( -1 );
	}
	if( byte_stream == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid byte stream.",
		 function );

		return( -1 );
	}
	if( byte_stream_size < 2 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: byte stream too small.",
		 function );

		return( -1 );
	}
	if( byte_stream_size > (size_t) SSIZE_MAX )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: byte stream size exceeds maximum.",
		 function );

		return( -1 );
	}
	byte_stream_copy_to_uint16_little_endian(
	 byte_stream,
	 extension_block->size );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: size\t\t\t: %" PRIu16 "\n",
		 function,
		 extension_block->size );
	}
#endif
	if( extension_block->size == 0 )
	{
		extension_block->data_size = 2;

		return( 1 );
	}
	extension_block->data_size = (size_t) extension_block->size;
	extension_block_data_size  = extension_block->data_size;

	if( ( extension_block->size < 10 )
	 && ( (size_t) extension_block->size > byte_stream_size ) )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_OUT_OF_BOUNDS,
		 "%s: invalid extension block size value out of bounds.",
		 function );

		goto on_error;
	}
	byte_stream               += 2;
	byte_stream_size          -= 2;
	extension_block_data_size -= 2;

	byte_stream_copy_to_uint16_little_endian(
	 byte_stream,
	 extension_block->version );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: version\t\t\t: %" PRIu16 "\n",
		 function,
		 extension_block->version );
	}
#endif
	byte_stream               += 2;
	byte_stream_size          -= 2;
	extension_block_data_size -= 2;

	byte_stream_copy_to_uint32_little_endian(
	 byte_stream,
	 extension_block->signature );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: signature\t\t: 0x%08" PRIx32 "\n",
		 function,
		 extension_block->signature );
	}
#endif
	byte_stream               += 4;
	byte_stream_size          -= 4;
	extension_block_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		if( libfguid_identifier_initialize(
		     &guid,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create GUID.",
			 function );

			goto on_error;
		}
		if( libfdatetime_fat_date_time_initialize(
		     &fat_date_time,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create FAT date time.",
			 function );

			goto on_error;
		}
	}
#endif
	switch( extension_block->signature )
	{
		case 0xbeef0003:
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				if( libfguid_identifier_copy_from_byte_stream(
				     guid,
				     byte_stream,
				     byte_stream_size,
				     LIBFGUID_ENDIAN_LITTLE,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy byte stream to GUID.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libfguid_identifier_copy_to_utf16_string(
					  guid,
					  (uint16_t *) guid_string,
					  48,
					  LIBFGUID_STRING_FORMAT_FLAG_USE_LOWER_CASE,
					  error );
#else
				result = libfguid_identifier_copy_to_utf8_string(
					  guid,
					  (uint8_t *) guid_string,
					  48,
					  LIBFGUID_STRING_FORMAT_FLAG_USE_LOWER_CASE,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy GUID to string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: class identifier\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
				 function,
				 guid_string );
				libcnotify_printf(
				 "%s: shell folder name\t: %s\n",
				 function,
				 libfwsi_folder_identifier_get_name(
				  byte_stream ) );
			}
#endif
			byte_stream               += 16;
			byte_stream_size          -= 16;
			extension_block_data_size -= 16;

			break;

		case 0xbeef0004:
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				if( libfdatetime_fat_date_time_copy_from_byte_stream(
				     fat_date_time,
				     byte_stream,
				     byte_stream_size,
				     LIBFDATETIME_ENDIAN_LITTLE,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy byte stream to FAT date time.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libfdatetime_fat_date_time_copy_to_utf16_string(
					  fat_date_time,
					  (uint16_t *) date_time_string,
					  32,
					  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
					  error );
#else
				result = libfdatetime_fat_date_time_copy_to_utf8_string(
					  fat_date_time,
					  (uint8_t *) date_time_string,
					  32,
					  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy FAT date time to string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: creation time\t\t: %" PRIs_LIBCSTRING_SYSTEM " UTC\n",
				 function,
				 date_time_string );
			}
#endif
			byte_stream               += 4;
			byte_stream_size          -= 4;
			extension_block_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				if( libfdatetime_fat_date_time_copy_from_byte_stream(
				     fat_date_time,
				     byte_stream,
				     byte_stream_size,
				     LIBFDATETIME_ENDIAN_LITTLE,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy byte stream to FAT date time.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libfdatetime_fat_date_time_copy_to_utf16_string(
					  fat_date_time,
					  (uint16_t *) date_time_string,
					  32,
					  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
					  error );
#else
				result = libfdatetime_fat_date_time_copy_to_utf8_string(
					  fat_date_time,
					  (uint8_t *) date_time_string,
					  32,
					  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy FAT date time to string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: access time\t\t: %" PRIs_LIBCSTRING_SYSTEM " UTC\n",
				 function,
				 date_time_string );
			}
#endif
			byte_stream               += 4;
			byte_stream_size          -= 4;
			extension_block_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				byte_stream_copy_to_uint16_little_endian(
				 byte_stream,
				 value_16bit );

				libcnotify_printf(
				 "%s: unknown1\t\t\t: 0x%04" PRIx16 "\n",
				 function,
				 value_16bit );
			}
#endif
			byte_stream               += 2;
			byte_stream_size          -= 2;
			extension_block_data_size -= 2;

			if( extension_block->version >= 7 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: unknown2\t\t\t: 0x%04" PRIx16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream               += 2;
				byte_stream_size          -= 2;
				extension_block_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: MFT entry index\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream               += 4;
				byte_stream_size          -= 4;
				extension_block_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: unknown3\t\t\t: %" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream               += 2;
				byte_stream_size          -= 2;
				extension_block_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: MFT sequence\t\t: %" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream               += 2;
				byte_stream_size          -= 2;
				extension_block_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: unknown4:\n",
					 function );
					libcnotify_print_data(
					 byte_stream,
					 8,
					 0 );
				}
#endif
				byte_stream               += 8;
				byte_stream_size          -= 8;
				extension_block_data_size -= 8;
			}
			byte_stream_copy_to_uint16_little_endian(
			 byte_stream,
			 long_string_size );

#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: long string size\t\t: %" PRIu16 "\n",
				 function,
				 long_string_size );
			}
#endif
			byte_stream               += 2;
			byte_stream_size          -= 2;
			extension_block_data_size -= 2;

			if( extension_block->version >= 8 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: unknown5:\n",
					 function );
					libcnotify_print_data(
					 byte_stream,
					 4,
					 0 );
				}
#endif
				byte_stream               += 4;
				byte_stream_size          -= 4;
				extension_block_data_size -= 4;
			}
			/* Determine the long name size
			 */
			for( string_size = 0;
			     ( string_size + 1 ) < extension_block_data_size - 2;
			     string_size += 2 )
			{
				if( ( byte_stream[ string_size ] == 0 )
				 && ( byte_stream[ string_size + 1 ] == 0 ) )
				{
					string_size += 2;

					break;
				}
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libuna_utf16_string_size_from_utf16_stream(
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  &value_string_size,
					  error );
#else
				result = libuna_utf8_string_size_from_utf16_stream(
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  &value_string_size,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to determine size of long name string.",
					 function );

					goto on_error;
				}
				if( ( value_string_size > (size_t) SSIZE_MAX )
				 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
					 "%s: invalid long name string size value exceeds maximum.",
					 function );

					goto on_error;
				}
				value_string = libcstring_system_string_allocate(
						value_string_size );

				if( value_string == NULL )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_MEMORY,
					 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
					 "%s: unable to create long name string.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libuna_utf16_string_copy_from_utf16_stream(
					  (libuna_utf16_character_t *) value_string,
					  value_string_size,
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  error );
#else
				result = libuna_utf8_string_copy_from_utf16_stream(
					  (libuna_utf8_character_t *) value_string,
					  value_string_size,
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
					 "%s: unable to set long name string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: long name\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
				 function,
				 value_string );

				memory_free(
				 value_string );

				value_string = NULL;
			}
#endif
			byte_stream               += string_size;
			byte_stream_size          -= string_size;
			extension_block_data_size -= string_size;

			if( long_string_size > 0 )
			{
				if( extension_block->version >= 7 )
				{
					/* Determine the localized long name size
					 */
					for( string_size = 0;
					     ( string_size + 1 ) < extension_block_data_size - 2;
					     string_size += 2 )
					{
						if( ( byte_stream[ string_size ] == 0 )
						 && ( byte_stream[ string_size + 1 ] == 0 ) )
						{
							string_size += 2;

							break;
						}
					}
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_size_from_utf16_stream(
							  byte_stream,
							  string_size,
							  LIBUNA_ENDIAN_LITTLE,
							  &value_string_size,
							  error );
#else
						result = libuna_utf8_string_size_from_utf16_stream(
							  byte_stream,
							  string_size,
							  LIBUNA_ENDIAN_LITTLE,
							  &value_string_size,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to determine size of localized long name string.",
							 function );

							goto on_error;
						}
						if( ( value_string_size > (size_t) SSIZE_MAX )
						 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
							 "%s: invalid localized long name string size value exceeds maximum.",
							 function );

							goto on_error;
						}
						value_string = libcstring_system_string_allocate(
								value_string_size );

						if( value_string == NULL )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_MEMORY,
							 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
							 "%s: unable to create localized long name string.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_copy_from_utf16_stream(
							  (libuna_utf16_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  string_size,
							  LIBUNA_ENDIAN_LITTLE,
							  error );
#else
						result = libuna_utf8_string_copy_from_utf16_stream(
							  (libuna_utf8_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  string_size,
							  LIBUNA_ENDIAN_LITTLE,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
							 "%s: unable to set localized long name string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: localized long name\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 value_string );

						memory_free(
						 value_string );

						value_string = NULL;
					}
#endif
				}
				else if( extension_block->version >= 3 )
				{
					/* Determine the localized long name size
					 */
					for( string_size = 0;
					     string_size < extension_block_data_size - 2;
					     string_size += 1 )
					{
						if( byte_stream[ string_size ] == 0 )
						{
							string_size += 1;

							break;
						}
					}
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_size_from_byte_stream(
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#else
						result = libuna_utf8_string_size_from_byte_stream(
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to determine size of localized long name string.",
							 function );

							goto on_error;
						}
						if( ( value_string_size > (size_t) SSIZE_MAX )
						 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
							 "%s: invalid localized long name string size value exceeds maximum.",
							 function );

							goto on_error;
						}
						value_string = libcstring_system_string_allocate(
								value_string_size );

						if( value_string == NULL )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_MEMORY,
							 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
							 "%s: unable to create localized long name string.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_copy_from_byte_stream(
							  (libuna_utf16_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  error );
#else
						result = libuna_utf8_string_copy_from_byte_stream(
							  (libuna_utf8_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
							 "%s: unable to set localized long name string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: localized long name\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 value_string );

						memory_free(
						 value_string );

						value_string = NULL;
					}
#endif
				}
				byte_stream               += string_size;
				byte_stream_size          -= string_size;
				extension_block_data_size -= string_size;
			}
			break;

		case 0xbeef0006:
			/* Determine the username size
			 */
			for( string_size = 0;
			     ( string_size + 1 ) < extension_block_data_size - 2;
			     string_size += 2 )
			{
				if( ( byte_stream[ string_size ] == 0 )
				 && ( byte_stream[ string_size + 1 ] == 0 ) )
				{
					string_size += 2;

					break;
				}
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libuna_utf16_string_size_from_utf16_stream(
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  &value_string_size,
					  error );
#else
				result = libuna_utf8_string_size_from_utf16_stream(
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  &value_string_size,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to determine size of username string.",
					 function );

					goto on_error;
				}
				if( ( value_string_size > (size_t) SSIZE_MAX )
				 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
					 "%s: invalid username string size value exceeds maximum.",
					 function );

					goto on_error;
				}
				value_string = libcstring_system_string_allocate(
						value_string_size );

				if( value_string == NULL )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_MEMORY,
					 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
					 "%s: unable to create username string.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libuna_utf16_string_copy_from_utf16_stream(
					  (libuna_utf16_character_t *) value_string,
					  value_string_size,
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  error );
#else
				result = libuna_utf8_string_copy_from_utf16_stream(
					  (libuna_utf8_character_t *) value_string,
					  value_string_size,
					  byte_stream,
					  string_size,
					  LIBUNA_ENDIAN_LITTLE,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
					 "%s: unable to set username string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: username\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
				 function,
				 value_string );

				memory_free(
				 value_string );

				value_string = NULL;
			}
#endif
			byte_stream               += string_size;
			byte_stream_size          -= string_size;
			extension_block_data_size -= string_size;

			break;

		case 0xbeef0014:
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				if( libfguid_identifier_copy_from_byte_stream(
				     guid,
				     byte_stream,
				     byte_stream_size,
				     LIBFGUID_ENDIAN_LITTLE,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy byte stream to GUID.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libfguid_identifier_copy_to_utf16_string(
					  guid,
					  (uint16_t *) guid_string,
					  48,
					  LIBFGUID_STRING_FORMAT_FLAG_USE_LOWER_CASE,
					  error );
#else
				result = libfguid_identifier_copy_to_utf8_string(
					  guid,
					  (uint8_t *) guid_string,
					  48,
					  LIBFGUID_STRING_FORMAT_FLAG_USE_LOWER_CASE,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy GUID to string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: class identifier\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
				 function,
				 guid_string );
			}
#endif
			result = memory_compare(
			          byte_stream,
			          curi_class_identifier,
			          16 );

			byte_stream               += 16;
			byte_stream_size          -= 16;
			extension_block_data_size -= 16;

			if( result == 0 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: data size\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream               += 4;
				byte_stream_size          -= 4;
				extension_block_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: unknown1:\n",
					 function );
					libcnotify_print_data(
					 byte_stream,
					 8,
					 0 );
				}
#endif
				byte_stream               += 8;
				byte_stream_size          -= 8;
				extension_block_data_size -= 8;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: unknown2\t\t\t: 0x%08" PRIx32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream               += 4;
				byte_stream_size          -= 4;
				extension_block_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: unknown3:\n",
					 function );
					libcnotify_print_data(
					 byte_stream,
					 12,
					 0 );
				}
#endif
				byte_stream               += 12;
				byte_stream_size          -= 12;
				extension_block_data_size -= 12;

				/* Property table
				 */
				byte_stream_copy_to_uint32_little_endian(
				 byte_stream,
				 number_of_properties );

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: number of properties\t: %" PRIu32 "\n",
					 function,
					 number_of_properties );
				}
#endif
				byte_stream               += 4;
				byte_stream_size          -= 4;
				extension_block_data_size -= 4;

				for( property_index = 0;
				     property_index < number_of_properties;
				     property_index++ )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 property_type );

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: property: %" PRIu32 " type\t\t: %" PRIu32 "\n",
						 function,
						 property_index,
						 property_type );
					}
#endif
					byte_stream               += 4;
					byte_stream_size          -= 4;
					extension_block_data_size -= 4;

					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 property_size );

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: property: %" PRIu32 " size\t\t: %" PRIu32 "\n",
						 function,
						 property_index,
						 property_size );
					}
#endif
					byte_stream               += 4;
					byte_stream_size          -= 4;
					extension_block_data_size -= 4;

/* TODO check if property_size value is sane */
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: property: %" PRIu32 " data:\n",
						 function,
						 property_index );
						libcnotify_print_data(
						 byte_stream,
						 property_size,
						 LIBCNOTIFY_PRINT_DATA_FLAG_GROUP_DATA );
					}
#endif
					byte_stream               += property_size;
					byte_stream_size          -= property_size;
					extension_block_data_size -= property_size;
				}
			}
			break;

		case 0xbeef0017:
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: unknown1:\n",
				 function );
				libcnotify_print_data(
				 byte_stream,
				 extension_block_data_size - 2,
				 LIBCNOTIFY_PRINT_DATA_FLAG_GROUP_DATA );
			}
#endif
			byte_stream               += extension_block_data_size - 2;
			byte_stream_size          -= extension_block_data_size - 2;
			extension_block_data_size -= extension_block_data_size - 2;

			break;

		default:
			break;
	}
	if( extension_block_data_size < 2 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_OUT_OF_BOUNDS,
		 "%s: invalid extension block data size value out of bounds.",
		 function );

		goto on_error;
	}
	if( extension_block_data_size > 2 )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libcnotify_verbose != 0 )
		{
			libcnotify_printf(
			 "%s: trailing data:\n",
			 function );
			libcnotify_print_data(
			 byte_stream,
			 extension_block_data_size - 2,
			 LIBCNOTIFY_PRINT_DATA_FLAG_GROUP_DATA );
		}
#endif
		byte_stream               += extension_block_data_size - 2;
		byte_stream_size          -= extension_block_data_size - 2;
		extension_block_data_size -= extension_block_data_size - 2;
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		byte_stream_copy_to_uint16_little_endian(
		 byte_stream,
		 value_16bit );

		libcnotify_printf(
		 "%s: version offset\t\t: %" PRIu16 "\n",
		 function,
		 value_16bit );
	}
#endif
	byte_stream               += 2;
	byte_stream_size          -= 2;
	extension_block_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		if( libfdatetime_fat_date_time_free(
		     &fat_date_time,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free FAT date time.",
			 function );

			goto on_error;
		}
		if( libfguid_identifier_free(
		     &guid,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free GUID.",
			 function );

			goto on_error;
		}
		libcnotify_printf(
		 "\n" );
	}
#endif
	return( 1 );

on_error:
#if defined( HAVE_DEBUG_OUTPUT )
	if( value_string != NULL )
	{
		memory_free(
		 value_string );
	}
	if( fat_date_time != NULL )
	{
		libfdatetime_fat_date_time_free(
		 &fat_date_time,
		 NULL );
	}
	if( guid != NULL )
	{
		libfguid_identifier_free(
		 &guid,
		 NULL );
	}
#endif
	return( -1 );
}

