/*
 * Windows Shell Item functions
 *
 * Copyright (c) 2010-2013, Joachim Metz <joachim.metz@gmail.com>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include "libfwsi_definitions.h"
#include "libfwsi_extension_block.h"
#include "libfwsi_file_attributes.h"
#include "libfwsi_file_entry_values.h"
#include "libfwsi_folder_identifier.h"
#include "libfwsi_item.h"
#include "libfwsi_libcerror.h"
#include "libfwsi_libcnotify.h"
#include "libfwsi_libcstring.h"
#include "libfwsi_libfdatetime.h"
#include "libfwsi_libfguid.h"
#include "libfwsi_libuna.h"
#include "libfwsi_root_folder_values.h"
#include "libfwsi_types.h"
#include "libfwsi_volume_entry_values.h"

/* Creates an item
 * Make sure the value item is referencing, is set to NULL
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_initialize(
     libfwsi_item_t **item,
     libcerror_error_t **error )
{
	libfwsi_internal_item_t *internal_item = NULL;
	static char *function                  = "libfwsi_item_initialize";

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	if( *item != NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid item value already set.",
		 function );

		return( -1 );
	}
	internal_item = memory_allocate_structure(
	                 libfwsi_internal_item_t );

	if( internal_item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create item.",
		 function );

		goto on_error;
	}
	if( memory_set(
	     internal_item,
	     0,
	     sizeof( libfwsi_internal_item_t ) ) == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear item.",
		 function );

		goto on_error;
	}
	*item = (libfwsi_item_t *) internal_item;

	return( 1 );

on_error:
	if( internal_item != NULL )
	{
		memory_free(
		 internal_item );
	}
	return( -1 );
}

/* Frees an item
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_free(
     libfwsi_item_t **item,
     libcerror_error_t **error )
{
	static char *function = "libfwsi_item_free";

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	if( *item != NULL )
	{
		/* The item is managed elsewhere
		 */
		*item = NULL;
	}
	return( 1 );
}

/* Frees an item
 * Returns 1 if successful or -1 on error
 */
int libfwsi_internal_item_free(
     libfwsi_internal_item_t **internal_item,
     libcerror_error_t **error )
{
	static char *function = "libfwsi_internal_item_free";
	int result            = 1;

	if( internal_item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	if( *internal_item != NULL )
	{
		if( ( *internal_item )->value != NULL )
		{
			if( ( *internal_item )->free_value != NULL )
			{
				if( ( *internal_item )->free_value(
				     &( ( *internal_item )->value ),
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
					 "%s: unable to free item value.",
					 function );

					result = -1;
				}
			}
		}
		memory_free(
		 *internal_item );

		*internal_item = NULL;
	}
	return( result );
}

/* Copies a shell item from a byte stream
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_copy_from_byte_stream(
     libfwsi_item_t *item,
     const uint8_t *byte_stream,
     size_t byte_stream_size,
     int ascii_codepage,
     libcerror_error_t **error )
{
        libfwsi_extension_block_t *extension_block         = NULL;
	libfwsi_internal_item_t *internal_item             = NULL;
	libfwsi_file_entry_values_t *file_entry_values     = NULL;
	libfwsi_root_folder_values_t *root_folder_values   = NULL;
	libfwsi_volume_entry_values_t *volume_entry_values = NULL;
	static char *function                              = "libfwsi_item_copy_from_byte_stream";
	size_t string_alignment_size                       = 0;
	size_t string_size                                 = 0;
	size_t shell_item_data_size                        = 0;
	ssize_t read_count                                 = 0;
	uint32_t signature                                 = 0;
	uint32_t string_data_size                          = 0;
	uint16_t data_size                                 = 0;
	uint16_t identifier_size                           = 0;
	uint16_t property_store_size                       = 0;
	uint8_t flags                                      = 0;
	uint8_t found_signature_match                      = 0;
	uint8_t has_trailing_extension_blocks              = 0;
	uint8_t in_pre_xp_format                           = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	libcstring_system_character_t date_time_string[ 32 ];
        libcstring_system_character_t guid_string[ 48 ];

	libcstring_system_character_t *value_string        = NULL;
	libfdatetime_fat_date_time_t *fat_date_time        = NULL;
	libfdatetime_filetime_t *filetime                  = NULL;
        libfguid_identifier_t *guid                        = NULL;
	size_t value_string_size                           = 0;
	uint32_t value_32bit                               = 0;
	uint16_t value_16bit                               = 0;
	int result                                         = 0;
#endif

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	internal_item = (libfwsi_internal_item_t *) item;

	if( byte_stream == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid byte stream.",
		 function );

		return( -1 );
	}
	if( byte_stream_size < 2 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: byte stream too small.",
		 function );

		return( -1 );
	}
	if( byte_stream_size > (size_t) SSIZE_MAX )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: byte stream size exceeds maximum.",
		 function );

		return( -1 );
	}
	byte_stream_copy_to_uint16_little_endian(
	 byte_stream,
	 internal_item->size );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: size\t\t\t\t: %" PRIu16 "\n",
		 function,
		 internal_item->size );
	}
#endif
	if( internal_item->size == 0 )
	{
		return( 1 );
	}
	shell_item_data_size = internal_item->size;

	if( ( internal_item->size < 4 )
	 && ( (size_t) internal_item->size > byte_stream_size ) )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_OUT_OF_BOUNDS,
		 "%s: invalid shell item size value out of bounds.",
		 function );

		goto on_error;
	}
	byte_stream          += 2;
	byte_stream_size     -= 2;
	shell_item_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: data:\n",
		 function );
		libcnotify_print_data(
		 byte_stream,
		 shell_item_data_size,
		 LIBCNOTIFY_PRINT_DATA_FLAG_GROUP_DATA );
	}
#endif
	internal_item->type = *byte_stream;

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: class type indicator\t\t: 0x%02" PRIx8 "\n",
		 function,
		 internal_item->type );
	}
#endif
	byte_stream          += 1;
	byte_stream_size     -= 1;
	shell_item_data_size -= 1;

	if( libfwsi_extension_block_initialize(
	     &extension_block,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
		 "%s: unable to create extension block.",
		 function );

		goto on_error;
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		if( libfguid_identifier_initialize(
		     &guid,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
			 "%s: unable to create GUID.",
			 function );

			goto on_error;
		}
	}
#endif
	switch( internal_item->type & 0x70 )
	{
		case 0x00:
			if( internal_item->size < 4 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
				 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
				 "%s: unsupported shell item size.",
				 function );

				goto on_error;
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: unknown0\t\t\t\t: 0x%02" PRIx8 "\n",
				 function,
				 *byte_stream );
			}
#endif
			byte_stream          += 1;
			byte_stream_size     -= 1;
			shell_item_data_size -= 1;

			if( internal_item->type != 0x00 )
			{
				break;
			}
			byte_stream_copy_to_uint32_little_endian(
			 byte_stream,
			 signature );

			switch( signature )
			{
				/* "AugM" */
				case 0x4d677541:
					found_signature_match         = 1;
					has_trailing_extension_blocks = 1;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: signature\t\t\t\t: 0x%08" PRIx32 "\n",
						 function,
						 signature );
					}
#endif
					byte_stream          += 4;
					byte_stream_size     -= 4;
					shell_item_data_size -= 4;

					break;

				/* "GFSI" */
				case 0x49534647:
					found_signature_match         = 1;
					has_trailing_extension_blocks = 0;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: signature\t\t\t\t: 0x%08" PRIx32 "\n",
						 function,
						 signature );
					}
#endif
					byte_stream          += 4;
					byte_stream_size     -= 4;
					shell_item_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						if( libfguid_identifier_copy_from_byte_stream(
						     guid,
						     byte_stream,
						     byte_stream_size,
						     LIBFGUID_ENDIAN_LITTLE,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy byte stream to GUID.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libfguid_identifier_copy_to_utf16_string(
							  guid,
							  (uint16_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#else
						result = libfguid_identifier_copy_to_utf8_string(
							  guid,
							  (uint8_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy GUID to string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: class identifier\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 guid_string );
					}
#endif
					byte_stream          += 16;
					byte_stream_size     -= 16;
					shell_item_data_size -= 16;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: unknown1:\n",
						 function );
						libcnotify_print_data(
						 byte_stream,
						 8,
						 0 );
					}
#endif
					byte_stream          += 8;
					byte_stream_size     -= 8;
					shell_item_data_size -= 8;

					break;

				case 0xffffff38:
					found_signature_match         = 1;
					has_trailing_extension_blocks = 0;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: signature\t\t\t\t: 0x%08" PRIx32 "\n",
						 function,
						 signature );
					}
#endif
					byte_stream          += 4;
					byte_stream_size     -= 4;
					shell_item_data_size -= 4;

/* TODO add support */
					break;

				default:
					found_signature_match         = 0;
					has_trailing_extension_blocks = 0;

					break;
			}
			if( found_signature_match == 0 )
			{
				byte_stream_copy_to_uint16_little_endian(
				 byte_stream,
				 data_size );

				if( data_size < shell_item_data_size )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: data size\t\t\t\t: %" PRIu16 "\n",
						 function,
						 data_size );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

					if( data_size != 0 )
					{
						if( data_size < 2 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_OUT_OF_BOUNDS,
							 "%s: invalid data size value out of bounds.",
							 function );

							goto on_error;
						}
						data_size -= 2;
					}
					if( data_size > 0 )
					{
						byte_stream_copy_to_uint32_little_endian(
						 byte_stream,
						 signature );

#if defined( HAVE_DEBUG_OUTPUT )
						if( libcnotify_verbose != 0 )
						{
							libcnotify_printf(
							 "%s: signature\t\t\t\t: 0x%08" PRIx32 "\n",
							 function,
							 signature );
						}
#endif
						byte_stream          += 4;
						byte_stream_size     -= 4;
						shell_item_data_size -= 4;

						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 property_store_size );

#if defined( HAVE_DEBUG_OUTPUT )
						if( libcnotify_verbose != 0 )
						{
							libcnotify_printf(
							 "%s: property store size\t\t\t: %" PRIu16 "\n",
							 function,
							 property_store_size );
						}
#endif
						byte_stream          += 2;
						byte_stream_size     -= 2;
						shell_item_data_size -= 2;

						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 identifier_size );

#if defined( HAVE_DEBUG_OUTPUT )
						if( libcnotify_verbose != 0 )
						{
							libcnotify_printf(
							 "%s: identifier size\t\t\t: %" PRIu16 "\n",
							 function,
							 identifier_size );
						}
#endif
						byte_stream          += 2;
						byte_stream_size     -= 2;
						shell_item_data_size -= 2;


						switch( signature )
						{
							case 0x10141981:
								if( identifier_size == 32 )
								{
/* TODO */
									found_signature_match = 0;
								}
								break;

							case 0x23febbee:
								if( identifier_size == 16 )
								{
									found_signature_match = 1;
#if defined( HAVE_DEBUG_OUTPUT )
									if( libcnotify_verbose != 0 )
									{
										if( libfguid_identifier_copy_from_byte_stream(
										     guid,
										     byte_stream,
										     byte_stream_size,
										     LIBFGUID_ENDIAN_LITTLE,
										     error ) != 1 )
										{
											libcerror_error_set(
											 error,
											 LIBCERROR_ERROR_DOMAIN_RUNTIME,
											 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
											 "%s: unable to copy byte stream to GUID.",
											 function );

											goto on_error;
										}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
										result = libfguid_identifier_copy_to_utf16_string(
											  guid,
											  (uint16_t *) guid_string,
											  48,
											  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
											  error );
#else
										result = libfguid_identifier_copy_to_utf8_string(
											  guid,
											  (uint8_t *) guid_string,
											  48,
											  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
											  error );
#endif
										if( result != 1 )
										{
											libcerror_error_set(
											 error,
											 LIBCERROR_ERROR_DOMAIN_RUNTIME,
											 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
											 "%s: unable to copy GUID to string.",
											 function );

											goto on_error;
										}
										libcnotify_printf(
										 "%s: shell folder identifier\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
										 function,
										 guid_string );
										libcnotify_printf(
										 "%s: shell folder name\t\t\t: %s\n",
										 function,
										 libfwsi_folder_identifier_get_name(
										  byte_stream ) );
									}
#endif
									byte_stream          += 16;
									byte_stream_size     -= 16;
									shell_item_data_size -= 16;
								}
								break;

							case 0x3b93afbb:
							case 0xbeebee00:
								if( identifier_size == 4 )
								{
									found_signature_match = 1;

#if defined( HAVE_DEBUG_OUTPUT )
									if( libcnotify_verbose != 0 )
									{
										byte_stream_copy_to_uint32_little_endian(
										 byte_stream,
										 value_32bit );

										libcnotify_printf(
										 "%s: identifier\t\t\t\t: 0x%08" PRIx32 "\n",
										 function,
										 value_32bit );
									}
#endif
									byte_stream          += 4;
									byte_stream_size     -= 4;
									shell_item_data_size -= 4;
								}
								break;

							default:
								found_signature_match = 0;
								break;
						}

					}
				}
				if( identifier_size > 0 )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: identifier data:\n",
						 function );
						libcnotify_print_data(
						 byte_stream,
						 identifier_size,
						 0 );
					}
#endif
					byte_stream          += identifier_size;
					byte_stream_size     -= identifier_size;
					shell_item_data_size -= identifier_size;
				}
				if( property_store_size > 0 )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: property store data:\n",
						 function );
						libcnotify_print_data(
						 byte_stream,
						 property_store_size,
						 0 );
					}
#endif
					byte_stream          += property_store_size;
					byte_stream_size     -= property_store_size;
					shell_item_data_size -= property_store_size;
				}
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: unknown1 size\t\t\t: %" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream          += 2;
				byte_stream_size     -= 2;
				shell_item_data_size -= 2;

				if( has_trailing_extension_blocks != 0 )
				{
					while( shell_item_data_size > 0 )
					{
						if( libfwsi_extension_block_copy_from_byte_stream(
						     extension_block,
						     byte_stream,
						     byte_stream_size,
						     ascii_codepage,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy byte stream to extension block.",
							 function );

							goto on_error;
						}
						byte_stream          += extension_block->data_size;
						byte_stream_size     -= extension_block->data_size;
						shell_item_data_size -= extension_block->data_size;
					}
				}
			}
			break;

		case 0x10:
			if( internal_item->type == 0x1f )
			{
				if( libfwsi_root_folder_values_initialize(
				     &root_folder_values,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
					 "%s: unable to create root folder values.",
					 function );

					goto on_error;
				}
				read_count = libfwsi_root_folder_values_read(
				              root_folder_values,
				              internal_item->type,
				              byte_stream,
				              shell_item_data_size,
				              error );

				if( read_count == -1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy byte stream to root folder values.",
					 function );

					goto on_error;
				}
				internal_item->value      = (intptr_t *) root_folder_values;
				internal_item->free_value = (int (*)(intptr_t **, libcerror_error_t **)) &libfwsi_root_folder_values_free;
				root_folder_values        = NULL;

				byte_stream          += read_count;
				byte_stream_size     -= read_count;
				shell_item_data_size -= read_count;

				if( shell_item_data_size > 2 )
				{
/* TODO add extention block to item */
					if( libfwsi_extension_block_copy_from_byte_stream(
					     extension_block,
					     byte_stream,
					     byte_stream_size,
					     ascii_codepage,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to extension block.",
						 function );

						goto on_error;
					}
					byte_stream          += extension_block->data_size;
					byte_stream_size     -= extension_block->data_size;
					shell_item_data_size -= extension_block->data_size;
				}
			}
			break;

		case 0x20:
			if( libfwsi_volume_entry_values_initialize(
			     &volume_entry_values,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create volume entry values.",
				 function );

				goto on_error;
			}
			read_count = libfwsi_volume_entry_values_read(
			              volume_entry_values,
			              internal_item->type,
			              byte_stream,
			              shell_item_data_size,
			              ascii_codepage,
			              error );

			if( read_count == -1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
				 "%s: unable to copy byte stream to volume entry values.",
				 function );

				goto on_error;
			}
			internal_item->value      = (intptr_t *) volume_entry_values;
			internal_item->free_value = (int (*)(intptr_t **, libcerror_error_t **)) &libfwsi_volume_entry_values_free;
			volume_entry_values       = NULL;

			byte_stream          += read_count;
			byte_stream_size     -= read_count;
			shell_item_data_size -= read_count;

			break;

		case 0x30:
			if( libfwsi_file_entry_values_initialize(
			     &file_entry_values,
			     error ) != 1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
				 "%s: unable to create file entry values.",
				 function );

				goto on_error;
			}
			read_count = libfwsi_file_entry_values_read(
			              file_entry_values,
			              internal_item->type,
			              byte_stream,
			              shell_item_data_size,
			              ascii_codepage,
			              error );

			if( read_count == -1 )
			{
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_RUNTIME,
				 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
				 "%s: unable to copy byte stream to file entry values.",
				 function );

				goto on_error;
			}
			internal_item->value      = (intptr_t *) file_entry_values;
			internal_item->free_value = (int (*)(intptr_t **, libcerror_error_t **)) &libfwsi_file_entry_values_free;
			in_pre_xp_format          = file_entry_values->in_pre_xp_format;
			file_entry_values         = NULL;

			byte_stream          += read_count;
			byte_stream_size     -= read_count;
			shell_item_data_size -= read_count;

			if( in_pre_xp_format != 0 )
			{
				if( ( internal_item->type & 0x80 ) != 0 )
				{
					if( shell_item_data_size == 16 )
					{
#if defined( HAVE_DEBUG_OUTPUT )
						if( libcnotify_verbose != 0 )
						{
							if( libfguid_identifier_copy_from_byte_stream(
							     guid,
							     byte_stream,
							     byte_stream_size,
							     LIBFGUID_ENDIAN_LITTLE,
							     error ) != 1 )
							{
								libcerror_error_set(
								 error,
								 LIBCERROR_ERROR_DOMAIN_RUNTIME,
								 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
								 "%s: unable to copy byte stream to GUID.",
								 function );

								goto on_error;
							}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
							result = libfguid_identifier_copy_to_utf16_string(
								  guid,
								  (uint16_t *) guid_string,
								  48,
								  LIBFGUID_STRING_FORMAT_FLAG_USE_LOWER_CASE,
								  error );
#else
							result = libfguid_identifier_copy_to_utf8_string(
								  guid,
								  (uint8_t *) guid_string,
								  48,
								  LIBFGUID_STRING_FORMAT_FLAG_USE_LOWER_CASE,
								  error );
#endif
							if( result != 1 )
							{
								libcerror_error_set(
								 error,
								 LIBCERROR_ERROR_DOMAIN_RUNTIME,
								 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
								 "%s: unable to copy GUID to string.",
								 function );

								goto on_error;
							}
							libcnotify_printf(
							 "%s: class identifier\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
							 function,
							 guid_string );
						}
#endif
						byte_stream          += 16;
						byte_stream_size     -= 16;
						shell_item_data_size -= 16;
					}
				}
			}
			else
			{
				/* The shell item can contain extension block 0xbeef0006 if there is more data.
				 */
				if( ( ( ( internal_item->type & 0x80 ) == 0 )
				  && ( shell_item_data_size > 8 ) )
				 || ( ( ( internal_item->type & 0x80 ) != 0 )
				  && ( shell_item_data_size > ( 28 + 8 ) ) ) )
				{
					if( libfwsi_extension_block_copy_from_byte_stream(
					     extension_block,
					     byte_stream,
					     byte_stream_size,
					     ascii_codepage,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to extension block.",
						 function );

						goto on_error;
					}
					byte_stream          += extension_block->data_size;
					byte_stream_size     -= extension_block->data_size;
					shell_item_data_size -= extension_block->data_size;
				}
				/* The shell item will contains extension block 0xbeef0003 if flag 0x80
				 * is set in the class type indicator.
				 */
				if( ( ( internal_item->type & 0x80 ) != 0 )
				 && ( shell_item_data_size >= 28 ) )
				{
					if( libfwsi_extension_block_copy_from_byte_stream(
					     extension_block,
					     byte_stream,
					     byte_stream_size,
					     ascii_codepage,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to extension block.",
						 function );

						goto on_error;
					}
					byte_stream          += extension_block->data_size;
					byte_stream_size     -= extension_block->data_size;
					shell_item_data_size -= extension_block->data_size;
				}
			}
			break;

		case 0x40:
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: unknown0\t\t\t\t: 0x%02" PRIx8 "\n",
				 function,
				 *byte_stream );
			}
#endif
			byte_stream          += 1;
			byte_stream_size     -= 1;
			shell_item_data_size -= 1;

			flags = *byte_stream;

#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: flags\t\t\t\t: 0x%02" PRIx8 "\n",
				 function,
				 flags );
			}
#endif
			byte_stream          += 1;
			byte_stream_size     -= 1;
			shell_item_data_size -= 1;

			/* Determine the network location
			 */
			for( string_size = 0;
			     string_size < shell_item_data_size;
			     string_size++ )
			{
				if( byte_stream[ string_size ] == 0 )
				{
					string_size++;

					break;
				}
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libuna_utf16_string_size_from_byte_stream(
					  byte_stream,
					  string_size,
					  ascii_codepage,
					  &value_string_size,
					  error );
#else
				result = libuna_utf8_string_size_from_byte_stream(
					  byte_stream,
					  string_size,
					  ascii_codepage,
					  &value_string_size,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to determine size of network location string.",
					 function );

					goto on_error;
				}
				if( ( value_string_size > (size_t) SSIZE_MAX )
				 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
					 "%s: invalid network location string size value exceeds maximum.",
					 function );

					goto on_error;
				}
				value_string = libcstring_system_string_allocate(
						value_string_size );

				if( value_string == NULL )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_MEMORY,
					 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
					 "%s: unable to create network location string.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libuna_utf16_string_copy_from_byte_stream(
					  (libuna_utf16_character_t *) value_string,
					  value_string_size,
					  byte_stream,
					  string_size,
					  ascii_codepage,
					  error );
#else
				result = libuna_utf8_string_copy_from_byte_stream(
					  (libuna_utf8_character_t *) value_string,
					  value_string_size,
					  byte_stream,
					  string_size,
					  ascii_codepage,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
					 "%s: unable to set network location string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: network location\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
				 function,
				 value_string );

				memory_free(
				 value_string );

				value_string = NULL;
			}
#endif
			byte_stream          += string_size;
			byte_stream_size     -= string_size;
			shell_item_data_size -= string_size;

			if( ( ( flags & 0x80 ) != 0 )
			 && ( shell_item_data_size > 0 ) )
			{
				/* Determine the network description
				 */
				for( string_size = 0;
				     string_size < shell_item_data_size;
				     string_size++ )
				{
					if( byte_stream[ string_size ] == 0 )
					{
						string_size++;

						break;
					}
				}
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_size_from_byte_stream(
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  &value_string_size,
						  error );
#else
					result = libuna_utf8_string_size_from_byte_stream(
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  &value_string_size,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to determine size of description string.",
						 function );

						goto on_error;
					}
					if( ( value_string_size > (size_t) SSIZE_MAX )
					 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
						 "%s: invalid description string size value exceeds maximum.",
						 function );

						goto on_error;
					}
					value_string = libcstring_system_string_allocate(
							value_string_size );

					if( value_string == NULL )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_MEMORY,
						 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
						 "%s: unable to create description string.",
						 function );

						goto on_error;
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_copy_from_byte_stream(
						  (libuna_utf16_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  error );
#else
					result = libuna_utf8_string_copy_from_byte_stream(
						  (libuna_utf8_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
						 "%s: unable to set description string.",
						 function );

						goto on_error;
					}
					libcnotify_printf(
					 "%s: network description\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
					 function,
					 value_string );

					memory_free(
					 value_string );

					value_string = NULL;
				}
#endif
				byte_stream          += string_size;
				byte_stream_size     -= string_size;
				shell_item_data_size -= string_size;
			}
			if( ( ( flags & 0x40 ) != 0 )
			 && ( shell_item_data_size > 0 ) )
			{
				/* Determine the network comments
				 */
				for( string_size = 0;
				     string_size < shell_item_data_size;
				     string_size++ )
				{
					if( byte_stream[ string_size ] == 0 )
					{
						string_size++;

						break;
					}
				}
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_size_from_byte_stream(
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  &value_string_size,
						  error );
#else
					result = libuna_utf8_string_size_from_byte_stream(
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  &value_string_size,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to determine size of comments string.",
						 function );

						goto on_error;
					}
					if( ( value_string_size > (size_t) SSIZE_MAX )
					 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
						 "%s: invalid comments string size value exceeds maximum.",
						 function );

						goto on_error;
					}
					value_string = libcstring_system_string_allocate(
							value_string_size );

					if( value_string == NULL )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_MEMORY,
						 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
						 "%s: unable to create comments string.",
						 function );

						goto on_error;
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_copy_from_byte_stream(
						  (libuna_utf16_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  error );
#else
					result = libuna_utf8_string_copy_from_byte_stream(
						  (libuna_utf8_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
						 "%s: unable to set comments string.",
						 function );

						goto on_error;
					}
					libcnotify_printf(
					 "%s: network comments\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
					 function,
					 value_string );

					memory_free(
					 value_string );

					value_string = NULL;
				}
#endif
				byte_stream          += string_size;
				byte_stream_size     -= string_size;
				shell_item_data_size -= string_size;
			}
			if( shell_item_data_size >= 2 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: unknown1\t\t\t\t: 0x%04" PRIx16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream          += 2;
				byte_stream_size     -= 2;
				shell_item_data_size -= 2;
			}
			break;

		case 0x60:
			flags = *byte_stream;

#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: flags\t\t\t\t: 0x%02" PRIx8 "\n",
				 function,
				 flags );
			}
#endif
			byte_stream          += 1;
			byte_stream_size     -= 1;
			shell_item_data_size -= 1;

			byte_stream_copy_to_uint16_little_endian(
			 byte_stream,
			 data_size );

#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: data size\t\t\t\t: %" PRIu16 "\n",
				 function,
				 value_16bit );
			}
#endif
			byte_stream          += 2;
			byte_stream_size     -= 2;
			shell_item_data_size -= 2;

			if( data_size != 0 )
			{
				if( ( data_size < 2 )
				 && ( data_size > shell_item_data_size ) )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_VALUE_OUT_OF_BOUNDS,
					 "%s: invalid data size value out of bounds.",
					 function );

					goto on_error;
				}
				data_size -= 2;
			}
			if( data_size >= 40 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: unknown1\t\t\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream          += 4;
				byte_stream_size     -= 4;
				shell_item_data_size -= 4;
				data_size            -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: unknown2\t\t\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream          += 4;
				byte_stream_size     -= 4;
				shell_item_data_size -= 4;
				data_size            -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libfdatetime_filetime_initialize(
				     &filetime,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
					 "%s: unable to create filetime.",
					 function );

					goto on_error;
				}
				if( libfdatetime_filetime_copy_from_byte_stream(
				     filetime,
				     byte_stream,
				     byte_stream_size,
				     LIBFDATETIME_ENDIAN_LITTLE,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy byte stream to filetime.",
					 function );

					goto on_error;
				}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
				result = libfdatetime_filetime_copy_to_utf16_string(
					  filetime,
					  (uint16_t *) date_time_string,
					  32,
					  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME_NANO_SECONDS,
					  error );
#else
				result = libfdatetime_filetime_copy_to_utf8_string(
					  filetime,
					  (uint8_t *) date_time_string,
					  32,
					  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME_NANO_SECONDS,
					  error );
#endif
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy filetime to string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: unknown3 time\t\t\t: %" PRIs_LIBCSTRING_SYSTEM " UTC\n",
				 function,
				 date_time_string );

				if( libfdatetime_filetime_free(
				     &filetime,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
					 "%s: unable to free filetime.",
					 function );

					goto on_error;
				}
#endif
				byte_stream          += 8;
				byte_stream_size     -= 8;
				shell_item_data_size -= 8;
				data_size            -= 8;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: unknown4\t\t\t\t: 0x%08" PRIx32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream          += 4;
				byte_stream_size     -= 4;
				shell_item_data_size -= 4;
				data_size            -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: unknown5:\n",
					 function );
					libcnotify_print_data(
					 byte_stream,
					 12,
					 0 );
				}
#endif
				byte_stream          += 12;
				byte_stream_size     -= 12;
				shell_item_data_size -= 12;
				data_size            -= 12;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: unknown6\t\t\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream          += 4;
				byte_stream_size     -= 4;
				shell_item_data_size -= 4;
				data_size            -= 4;

				byte_stream_copy_to_uint32_little_endian(
				 byte_stream,
				 string_data_size );

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: string1 data size\t\t\t: %" PRIu32 "\n",
					 function,
					 string_data_size );
				}
#endif
				byte_stream          += 4;
				byte_stream_size     -= 4;
				shell_item_data_size -= 4;
				data_size            -= 4;
			}
/* TODO complete parsing of the 3 strings */
			if( data_size > 0 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: unknown data:\n",
					 function );
					libcnotify_print_data(
					 byte_stream,
					 data_size,
					 0 );
				}
#endif
				byte_stream          += data_size;
				byte_stream_size     -= data_size;
				shell_item_data_size -= data_size;
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				byte_stream_copy_to_uint16_little_endian(
				 byte_stream,
				 value_16bit );

				libcnotify_printf(
				 "%s: unknown7\t\t\t\t: %" PRIu16 "\n",
				 function,
				 value_16bit );
			}
#endif
			byte_stream          += 2;
			byte_stream_size     -= 2;
			shell_item_data_size -= 2;

			/* Determine the URI size
			 */
			if( ( flags & 0x80 ) != 0 )
			{
				for( string_size = 0;
				     ( string_size + 1 ) < shell_item_data_size;
				     string_size += 2 )
				{
					if( ( byte_stream[ string_size ] == 0 )
					 && ( byte_stream[ string_size + 1 ] == 0 ) )
					{
						string_size += 2;

						break;
					}
				}
			}
			else
			{
				for( string_size = 0;
				     string_size < shell_item_data_size;
				     string_size++ )
				{
					if( byte_stream[ string_size ] == 0 )
					{
						string_size++;

						break;
					}
				}
			}
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				if( ( flags & 0x80 ) != 0 )
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_size_from_utf16_stream(
						  byte_stream,
						  string_size,
						  LIBUNA_ENDIAN_LITTLE,
						  &value_string_size,
						  error );
#else
					result = libuna_utf8_string_size_from_utf16_stream(
						  byte_stream,
						  string_size,
						  LIBUNA_ENDIAN_LITTLE,
						  &value_string_size,
						  error );
#endif
				}
				else
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_size_from_byte_stream(
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  &value_string_size,
						  error );
#else
					result = libuna_utf8_string_size_from_byte_stream(
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  &value_string_size,
						  error );
#endif
				}
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
					 "%s: unable to determine size of URI string.",
					 function );

					goto on_error;
				}
				if( ( value_string_size > (size_t) SSIZE_MAX )
				 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
					 "%s: invalid URI string size value exceeds maximum.",
					 function );

					goto on_error;
				}
				value_string = libcstring_system_string_allocate(
						value_string_size );

				if( value_string == NULL )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_MEMORY,
					 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
					 "%s: unable to create URI string.",
					 function );

					goto on_error;
				}
				if( ( flags & 0x80 ) != 0 )
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_copy_from_utf16_stream(
						  (libuna_utf16_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  LIBUNA_ENDIAN_LITTLE,
						  error );
#else
					result = libuna_utf8_string_copy_from_utf16_stream(
						  (libuna_utf8_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  LIBUNA_ENDIAN_LITTLE,
						  error );
#endif
				}
				else
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_copy_from_byte_stream(
						  (libuna_utf16_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  error );
#else
					result = libuna_utf8_string_copy_from_byte_stream(
						  (libuna_utf8_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  string_size,
						  ascii_codepage,
						  error );
#endif
				}
				if( result != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
					 "%s: unable to set URI string.",
					 function );

					goto on_error;
				}
				libcnotify_printf(
				 "%s: URI\t\t\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
				 function,
				 value_string );

				memory_free(
				 value_string );

				value_string = NULL;
			}
#endif
			byte_stream          += string_size;
			byte_stream_size     -= string_size;
			shell_item_data_size -= string_size;

/* TODO value likely controlled by flags */
			if( shell_item_data_size > 0 )
			{
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: unknown9\t\t\t\t: %" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream          += 2;
				byte_stream_size     -= 2;
				shell_item_data_size -= 2;
			}
			if( shell_item_data_size > 0 )
			{
				if( libfwsi_extension_block_copy_from_byte_stream(
				     extension_block,
				     byte_stream,
				     byte_stream_size,
				     ascii_codepage,
				     error ) != 1 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_RUNTIME,
					 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
					 "%s: unable to copy byte stream to extension block.",
					 function );

					goto on_error;
				}
				byte_stream          += extension_block->data_size;
				byte_stream_size     -= extension_block->data_size;
				shell_item_data_size -= extension_block->data_size;
			}
			break;

		case 0x70:
#if defined( HAVE_DEBUG_OUTPUT )
			if( libcnotify_verbose != 0 )
			{
				libcnotify_printf(
				 "%s: unknown0\t\t\t\t: 0x%02" PRIx8 "\n",
				 function,
				 *byte_stream );
			}
#endif
			byte_stream          += 1;
			byte_stream_size     -= 1;
			shell_item_data_size -= 1;

			switch( internal_item->type )
			{
				case 0x71:
					if( internal_item->size != 30 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
						 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
						 "%s: unsupported shell item size.",
						 function );

						goto on_error;
					}
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: unknown1:\n",
						 function );
						libcnotify_print_data(
						 byte_stream,
						 10,
						 0 );
					}
#endif
					byte_stream          += 10;
					byte_stream_size     -= 10;
					shell_item_data_size -= 10;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						if( libfguid_identifier_copy_from_byte_stream(
						     guid,
						     byte_stream,
						     byte_stream_size,
						     LIBFGUID_ENDIAN_LITTLE,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy byte stream to GUID.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libfguid_identifier_copy_to_utf16_string(
							  guid,
							  (uint16_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#else
						result = libfguid_identifier_copy_to_utf8_string(
							  guid,
							  (uint8_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy GUID to string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: shell folder identifier\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 guid_string );
						libcnotify_printf(
						 "%s: shell folder name\t\t\t: %s\n",
						 function,
						 libfwsi_folder_identifier_get_name(
						  byte_stream ) );
					}
#endif
					byte_stream          += 16;
					byte_stream_size     -= 16;
					shell_item_data_size -= 16;

					break;

				case 0x74:
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: unknown1\t\t\t\t: 0x%04" PRIx16 "\n",
						 function,
						 value_16bit );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint32_little_endian(
						 byte_stream,
						 value_32bit );

						libcnotify_printf(
						 "%s: signature\t\t\t\t: 0x%04" PRIx32 "\n",
						 function,
						 value_32bit );
					}
#endif
					byte_stream          += 4;
					byte_stream_size     -= 4;
					shell_item_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: unknown2\t\t\t\t: 0x%04" PRIx16 "\n",
						 function,
						 value_16bit );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: unknown3\t\t\t\t: 0x%04" PRIx16 "\n",
						 function,
						 value_16bit );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint32_little_endian(
						 byte_stream,
						 value_32bit );

						libcnotify_printf(
						 "%s: file size\t\t\t\t: %" PRIu32 "\n",
						 function,
						 value_32bit );
					}
#endif
					byte_stream          += 4;
					byte_stream_size     -= 4;
					shell_item_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						if( libfdatetime_fat_date_time_initialize(
						     &fat_date_time,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
							 "%s: unable to create FAT date time.",
							 function );

							goto on_error;
						}
						if( libfdatetime_fat_date_time_copy_from_byte_stream(
						     fat_date_time,
						     byte_stream,
						     byte_stream_size,
						     LIBFDATETIME_ENDIAN_LITTLE,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy byte stream to FAT date time.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libfdatetime_fat_date_time_copy_to_utf16_string(
							  fat_date_time,
							  (uint16_t *) date_time_string,
							  32,
							  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
							  error );
#else
						result = libfdatetime_fat_date_time_copy_to_utf8_string(
							  fat_date_time,
							  (uint8_t *) date_time_string,
							  32,
							  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy FAT date time to string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: modification time\t\t\t: %" PRIs_LIBCSTRING_SYSTEM " UTC\n",
						 function,
						 date_time_string );

						if( libfdatetime_fat_date_time_free(
						     &fat_date_time,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
							 "%s: unable to free FAT date time.",
							 function );

							goto on_error;
						}
					}
#endif
					byte_stream          += 4;
					byte_stream_size     -= 4;
					shell_item_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: file attribute flags\t\t: 0x%04" PRIx16 "\n",
						 function,
						 value_16bit );
						libfwsi_file_attributes_print(
						 value_16bit );
						libcnotify_printf(
						 "\n" );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

					for( string_size = 0;
					     string_size < shell_item_data_size;
					     string_size++ )
					{
						if( byte_stream[ string_size ] == 0 )
						{
							string_size++;

							break;
						}
					}
					string_alignment_size = string_size % 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_size_from_byte_stream(
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#else
						result = libuna_utf8_string_size_from_byte_stream(
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to determine size of primary name string.",
							 function );

							goto on_error;
						}
						if( ( value_string_size > (size_t) SSIZE_MAX )
						 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
							 "%s: invalid primary name string size value exceeds maximum.",
							 function );

							goto on_error;
						}
						value_string = libcstring_system_string_allocate(
								value_string_size );

						if( value_string == NULL )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_MEMORY,
							 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
							 "%s: unable to create primary name string.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_copy_from_byte_stream(
							  (libuna_utf16_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  error );
#else
						result = libuna_utf8_string_copy_from_byte_stream(
							  (libuna_utf8_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  string_size,
							  ascii_codepage,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
							 "%s: unable to set primary name string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: primary name\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 value_string );

						memory_free(
						 value_string );

						value_string = NULL;
					}
#endif
					byte_stream          += string_size + string_alignment_size;
					byte_stream_size     -= string_size + string_alignment_size;
					shell_item_data_size -= string_size + string_alignment_size;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: unknown5\t\t\t\t: 0x%04" PRIx16 "\n",
						 function,
						 value_16bit );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						if( libfguid_identifier_copy_from_byte_stream(
						     guid,
						     byte_stream,
						     byte_stream_size,
						     LIBFGUID_ENDIAN_LITTLE,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy byte stream to GUID.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libfguid_identifier_copy_to_utf16_string(
							  guid,
							  (uint16_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#else
						result = libfguid_identifier_copy_to_utf8_string(
							  guid,
							  (uint8_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy GUID to string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: delagate item class identifier\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 guid_string );
					}
#endif
					byte_stream          += 16;
					byte_stream_size     -= 16;
					shell_item_data_size -= 16;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						if( libfguid_identifier_copy_from_byte_stream(
						     guid,
						     byte_stream,
						     byte_stream_size,
						     LIBFGUID_ENDIAN_LITTLE,
						     error ) != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy byte stream to GUID.",
							 function );

							goto on_error;
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libfguid_identifier_copy_to_utf16_string(
							  guid,
							  (uint16_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#else
						result = libfguid_identifier_copy_to_utf8_string(
							  guid,
							  (uint8_t *) guid_string,
							  48,
							  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
							 "%s: unable to copy GUID to string.",
							 function );

							goto on_error;
						}
						libcnotify_printf(
						 "%s: item class identifier\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 guid_string );
						libcnotify_printf(
						 "%s: item class name\t\t\t: %s\n",
						 function,
						 libfwsi_folder_identifier_get_name(
						  byte_stream ) );
					}
#endif
					byte_stream          += 16;
					byte_stream_size     -= 16;
					shell_item_data_size -= 16;

					if( libfwsi_extension_block_copy_from_byte_stream(
					     extension_block,
					     byte_stream,
					     byte_stream_size,
					     ascii_codepage,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to extension block.",
						 function );

						goto on_error;
					}
					byte_stream          += extension_block->data_size;
					byte_stream_size     -= extension_block->data_size;
					shell_item_data_size -= extension_block->data_size;

					break;
			}
			break;

		default:
			break;
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		if( shell_item_data_size > 0 )
		{
			libcnotify_printf(
			 "%s: trailing data:\n",
			 function );
			libcnotify_print_data(
			 byte_stream,
			 shell_item_data_size,
			 LIBCNOTIFY_PRINT_DATA_FLAG_GROUP_DATA );
		}
		if( libfguid_identifier_free(
		     &guid,
		     error ) != 1 )
		{
			libcerror_error_set(
			 error,
			 LIBCERROR_ERROR_DOMAIN_RUNTIME,
			 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
			 "%s: unable to free GUID.",
			 function );

			goto on_error;
		}
	}
#endif
	if( libfwsi_extension_block_free(
	     &extension_block,
	     error ) != 1 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
		 "%s: unable to free extension block.",
		 function );

		goto on_error;
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "\n" );
	}
#endif
	return( 1 );

on_error:
#if defined( HAVE_DEBUG_OUTPUT )
	if( value_string != NULL )
	{
		memory_free(
		 value_string );
	}
	if( filetime != NULL )
	{
		libfdatetime_filetime_free(
		 &filetime,
		 NULL );
	}
	if( fat_date_time != NULL )
	{
		libfdatetime_fat_date_time_free(
		 &fat_date_time,
		 NULL );
	}
	if( guid != NULL )
	{
		libfguid_identifier_free(
		 &guid,
		 NULL );
	}
#endif
	if( file_entry_values != NULL )
	{
		libfwsi_file_entry_values_free(
		 &file_entry_values,
		 NULL );
	}
	if( root_folder_values != NULL )
	{
		libfwsi_root_folder_values_free(
		 &root_folder_values,
		 NULL );
	}
	if( extension_block != NULL )
	{
		libfwsi_extension_block_free(
		 &extension_block,
		 NULL );
	}
	return( -1 );
}

/* Retrieved the item type
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_get_type(
     libfwsi_item_t *item,
     uint8_t *item_type,
     libcerror_error_t **error )
{
	libfwsi_internal_item_t *internal_item = NULL;
	static char *function                  = "libfwsi_item_get_type";

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	internal_item = (libfwsi_internal_item_t *) item;

	if( item_type == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item type.",
		 function );

		return( -1 );
	}
	*item_type = internal_item->type;

	return( 1 );
}

